/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base;

import java.util.ArrayList;
import java.util.List;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures.IndicatorType;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IClickController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.collections.ObservableList;
import javafx.scene.Node;
import javafx.scene.control.Menu;
import javafx.scene.control.MenuItem;

/**
 * Default implementation for diagram controller, which provides a {@code Display...} context menu
 * with the options to enable and disable link highlighting and interactive area shading.
 */
public class DefaultDiagramController extends ControllerBase {
	/** Flag for ignoring mouse click event after dragging gesture. */
	private boolean ignoreNextMouseClick = false;

	/** Constructor. */
	public DefaultDiagramController(IMVCBundle mvcb) {
		super(mvcb);
	}

	/** Creates the {@link MenuItem} to enable/disable link highlighting. */
	private MenuItem createLinkHighlightingMenuItem(DiagramViewerFeatures features) {
		MenuItem linkHighLightingMenuItem;
		if(features.isLinkHighlightingEnabled()) {
			linkHighLightingMenuItem = new MenuItem("Disable Link Highlighting");
			linkHighLightingMenuItem.setOnAction(evt -> {
				features.setLinkHighlightingEnabled(false);
			});
		} else {
			linkHighLightingMenuItem = new MenuItem("Enable Link Highlighting");
			linkHighLightingMenuItem.setOnAction(evt -> {
				features.setLinkHighlightingEnabled(true);
			});
		}
		return linkHighLightingMenuItem;
	}

	/** Creates a {@link MenuItem} to enable/disable interaction shading. */
	private MenuItem createInteractionAreaShadingMenuItem(DiagramViewerFeatures features) {
		MenuItem interactionAreaShadingMenuItem;
		if(features.isInteractionAreaShadingEnabled()) {
			interactionAreaShadingMenuItem = new MenuItem("Disable Interactive Area Shading");
			interactionAreaShadingMenuItem.setOnAction(evt -> {
				features.setInteractionAreaShadingEnabled(false);
			});
		} else {
			interactionAreaShadingMenuItem = new MenuItem("Enable Interactive Area Shading");
			interactionAreaShadingMenuItem.setOnAction(evt -> {
				features.setInteractionAreaShadingEnabled(true);
			});
		}
		return interactionAreaShadingMenuItem;
	}

	/**
	 * Creates a {@link MenuItem} to hide/show the grid (returns {@code null} in case the diagram
	 * does not have a grid).
	 */
	private MenuItem createGridMenuItem(DiagramViewerFeatures features) {
		MenuItem gridMenuItem = null;
		if(features.getIndicatorType() == IndicatorType.INVISIBLE) {
			gridMenuItem = new MenuItem("Show Grid");
			gridMenuItem.setOnAction(evt -> {
				features.setIndicatorTypeToDefault();
			});
		} else {
			gridMenuItem = new MenuItem("Hide Grid");
			gridMenuItem.setOnAction(evt -> {
				features.setIndicatorType(IndicatorType.INVISIBLE);
			});
		}
		return gridMenuItem;
	}

	/** Creates a {@link MenuItem} to zoom in. */
	private MenuItem createZoomInMenuItem(DiagramViewerFeatures features) {
		MenuItem zoomInMenuItem = new MenuItem("Zoom in");
		zoomInMenuItem.setOnAction(evt -> {
			features.zoomIn();
		});
		return zoomInMenuItem;
	}

	/** Creates a {@link MenuItem} to zoom out. */
	private MenuItem createZoomOutMenuItem(DiagramViewerFeatures features) {
		MenuItem zoomOutMenuItem = new MenuItem("Zoom out");
		zoomOutMenuItem.setOnAction(evt -> {
			features.zoomOut();
		});
		return zoomOutMenuItem;
	}

	/** Creates a {@link MenuItem} to set the zoom factor. */
	private MenuItem createZoomFactorMenuItem(DiagramViewerFeatures features, int zoomFactorIndex,
			String prefix) {
		MenuItem zoomFactorMenuItem = null;
		if(zoomFactorIndex >= 0 && zoomFactorIndex < features.getZoomFactors().length) {
			double zoomFactor = features.getZoomFactors()[zoomFactorIndex];
			zoomFactorMenuItem = new MenuItem(prefix + zoomFactor * 100 + " %");
			zoomFactorMenuItem.setOnAction(evt -> {
				features.setZoomFactorIndex(zoomFactorIndex);
			});
		}
		return zoomFactorMenuItem;
	}

	/** {@inheritDoc} */
	@Override
	public List<MenuItem> contextMenuContributions(Node node, DiagramCoordinate diagramLocation) {
		List<MenuItem> rval = new ArrayList<>();
		Menu gridMenu = new Menu("Display ...");
		ObservableList<MenuItem> gridMenuItems = gridMenu.getItems();
		DiagramViewerFeatures features = getViewer().getFeatures();

		gridMenuItems.add(createLinkHighlightingMenuItem(features));
		gridMenuItems.add(createInteractionAreaShadingMenuItem(features));
		MenuItem gridMenuItem = createGridMenuItem(features);
		if(gridMenuItem != null) {
			gridMenuItems.add(gridMenuItem);
		}
		rval.add(gridMenu);

		Menu zoomMenu = new Menu("Zoom ...");
		ObservableList<MenuItem> zoomMenuItems = zoomMenu.getItems();
		double[] zoomFactors = features.getZoomFactors();
		int normalZoomIndex = -1;
		for(int i = 0; i < zoomFactors.length; ++i) {
			MenuItem zoomFactorMenuItem = createZoomFactorMenuItem(features, i, "");
			if(zoomFactorMenuItem != null) {
				zoomMenuItems.add(zoomFactorMenuItem);
			}
			if(zoomFactors[i] == 1.0) {
				normalZoomIndex = i;
			}
		}
		rval.add(createZoomInMenuItem(features));
		rval.add(createZoomOutMenuItem(features));
		if(normalZoomIndex != -1) {
			rval.add(createZoomFactorMenuItem(features, normalZoomIndex, "Zoom "));
		}
		rval.add(zoomMenu);

		return rval;
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(gesture == EDragGesture.SELECTION) {
			return dragSelectionRectangleController;
		}
		return super.getDragController(gesture, node, diagramLocation);
	}

	/** {@link IDragController} for the selection rectangle. */
	private final IDragController dragSelectionRectangleController = new DragControllerBase() {
		@Override
		protected void safeDragStarted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			getViewer().startSelectionFeedback(node, diagramLocation);
			ignoreNextMouseClick = true;
		}

		@Override
		protected void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			getViewer().updateSelectionFeedback(node, diagramLocation);
		}

		@Override
		protected Change safeDragCompleted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			getViewer().terminateSelectionFeedback(node, diagramLocation);
			return null;
		}
	};

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		if(ignoreNextMouseClick) {
			ignoreNextMouseClick = false;
			return null;
		}
		return super.getClickController(node, diagramLocation);
	}
}
