/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.ui.utils;

import static java.lang.String.join;
import static javafx.scene.control.Alert.AlertType.INFORMATION;
import static org.conqat.ide.commons.ui.dialog.MessageUtils.askQuestion;
import static org.conqat.ide.commons.ui.dialog.MessageUtils.showInfo;
import static org.conqat.ide.commons.ui.dialog.MessageUtils.showWarning;
import static org.conqat.ide.commons.ui.ui.WorkbenchUtils.getActiveWorkbenchWindow;
import static org.fortiss.tooling.common.ui.javafx.util.GraphicUtils.getFXImage;
import static org.fortiss.tooling.ext.reuse.storage.ReuseLibraryStorageManager.REUSE_PROJECT_DIR;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getAllReuseSpecs;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getElementInsideLibrary;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getExternalReferencesOfElement;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getLibraryIDNameFormat;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementName;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.swt.widgets.Shell;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.ext.reuse.model.ReuseElementSpec;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.ui.ToolingReuseUIActivator;
import org.fortiss.tooling.ext.reuse.ui.dialog.SelectLibDialog;

import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.DialogPane;
import javafx.scene.control.TextArea;
import javafx.scene.image.Image;
import javafx.stage.Stage;

/**
 * Utility methods for {@link ReuseLibrary}s' UI.
 *
 * @author bergemann
 */
public class ReuseLibraryUIUtils {

	/** Standard width preference for dialog windows used in the reuse context. */
	private final static Double DIALOG_WINDOW_WIDTH_PREF = 700.0;

	/** Standard height preference for dialog windows used in the reuse context. */
	private final static Double DIALOG_WINDOW_HEIGHT_PREF = 500.0;

	/**
	 * Returns the selected {@link ReuseElementSpec} by user from all
	 * {@link ReuseElementSpec}s of the given element. Returns null if no
	 * {@link ReuseElementSpec} could be selected.
	 *
	 * @param element
	 *            The target reuse element (source of the
	 *            {@link ReuseElementSpec})
	 * @param dialogTitle
	 *            The title for the dialog with the user
	 * @return The {@link ReuseElementSpec} selected by the user
	 */
	public static ReuseElementSpec selectCorrectReuseSpec(EObject element, String dialogTitle) {
		if(element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs(element);
			if(specs.size() == 1) {
				// get(0) is safe due to size check.
				return specs.get(0);
			} else if(specs.size() > 1) {
				// Ask for selection.
				List<String> referencedLibrarySelection = new ArrayList<String>();
				for(ReuseElementSpec spec : specs) {
					referencedLibrarySelection.add(getLibraryIDNameFormat(spec.getSourceLibUUID(),
							spec.getSourceLibName()) + " with reference to element '" +
							spec.getElementName() + "'");
				}
				final Shell shell = getActiveWorkbenchWindow().getShell();
				final SelectLibDialog selectLibraryDialog = new SelectLibDialog(shell, dialogTitle);
				selectLibraryDialog.setLibraryNameList(referencedLibrarySelection);
				selectLibraryDialog.open();

				if(selectLibraryDialog.goAhead()) {
					return specs.get(selectLibraryDialog.getSelectionIndex());
				}
			}
		}
		return null;
	}

	/**
	 * Returns whether the user wants to keep all the additional
	 * {@link ReuseElementSpec}s in the given reuse {@link EObject} besides the
	 * given main {@link ReuseElementSpec} or wants to override/delete all of them.
	 * If no additional {@link ReuseElementSpec} was found, it will return true.
	 *
	 * @param element
	 *            The target reuse element (containing the library
	 *            references)
	 * @param mainReference
	 *            The main library reference to which the others are
	 *            additional
	 * @param dialogTitle
	 *            The title for the dialog with the user
	 * @return True if (possible) additional library references should be kept or
	 *         false if they should be overridden
	 */
	public static boolean isKeepingAdditionalReferencesRequested(EObject element,
			ReuseElementSpec mainReference, String dialogTitle) {
		if(element instanceof IModelElement) {
			List<ReuseElementSpec> specs = getAllReuseSpecs(element);
			// Only if we have any additional references/reuse specifications, we need to
			// ask (otherwise keeping is false, because "not existing refs/specs" do not be
			// need to kept).
			if(specs.size() > 1) {
				List<String> additionalLibraryReferences = new ArrayList<String>();
				for(ReuseElementSpec spec : specs) {
					if(spec != mainReference) {
						additionalLibraryReferences.add("- Library '" +
								getLibraryIDNameFormat(spec.getSourceLibUUID(),
										spec.getSourceLibName()) +
								"' with '" + spec.getElementName() +
								"' as element name in the library");
					}
				}

				List<String> textLines = new ArrayList<String>();
				textLines.add(
						"The following additional library references were found in the selected element:");
				textLines.addAll(additionalLibraryReferences);
				textLines.add("\nDo you want to keep them after the update?");
				textLines.add(
						"- Yes: All additional (old) references will be kept together with the one of the current update.");
				textLines.add(
						"- No: Then, only the library reference of the current update will further exist.");
				return askQuestion(dialogTitle, join("\n", textLines));
			}
		}
		return false;
	}

	/**
	 * Returns the original element inside the given {@link ReuseLibrary} referenced
	 * by the given {@link ReuseElementSpec}. Also, displays warning (with
	 * dialogTitle as header) if it could not be found, and then returns null.
	 *
	 * @param library
	 *            The target/source reuse library
	 * @param referencingSpec
	 *            The reuse specification that reference the requested
	 *            reuse element
	 * @param dialogTitle
	 *            The title for the dialog
	 * @return The requested reuse element if it could be found, otherwise null
	 */
	public static EObject checkForElementInLibrary(ReuseLibrary library,
			ReuseElementSpec referencingSpec, String dialogTitle) {
		String targetUuid = referencingSpec.getElementUUID();
		EObject originalElementInLibrary = getElementInsideLibrary(library, targetUuid);
		if(originalElementInLibrary == null) {
			warnOfMissingReuseElem(referencingSpec, dialogTitle);
		}
		return originalElementInLibrary;
	}

	/**
	 * Displays a warning message to the user about a missing reuse element (with
	 * extracted properties from the given {@link ReuseElementSpec} and title).
	 *
	 * @param referencingSpec
	 *            The reuse specification that has the information about
	 *            the missing reuse element
	 * @param dialogTitle
	 *            The title for the dialog
	 */
	public static void warnOfMissingReuseElem(ReuseElementSpec referencingSpec,
			String dialogTitle) {
		List<String> textLines = new ArrayList<String>();
		textLines.add("The referenced element does not exist in the referenced library.");
		textLines.add("\nSearched for element UUID: " + referencingSpec.getElementUUID());
		textLines.add("in library: " + getLibraryIDNameFormat(referencingSpec.getSourceLibUUID(),
				referencingSpec.getSourceLibName()));
		showWarning(dialogTitle + " - Missing Element", join("\n", textLines));
	}

	/**
	 * Displays a warning message to the user about a missing {@link ReuseLibrary}
	 * (with extracted properties from the given {@link ReuseElementSpec} and
	 * title).
	 *
	 * @param referencingSpec
	 *            The reuse specification that has the information about
	 *            the missing reuse library
	 * @param dialogTitle
	 *            The title for the dialog
	 */
	public static void warnOfMissingLib(ReuseElementSpec referencingSpec, String dialogTitle) {
		List<String> textLines = new ArrayList<String>();
		textLines.add("The selected library does not exist in the library directory.");
		textLines.add("\nPlease add it there and repeat.");
		textLines.add("\nMissing library (UUID and name): " + getLibraryIDNameFormat(
				referencingSpec.getSourceLibUUID(), referencingSpec.getSourceLibName()));
		textLines.add("\nLibrary directory location: " + REUSE_PROJECT_DIR.getAbsolutePath());
		showWarning(dialogTitle + " - Missing Library", join("\n", textLines));
	}

	/**
	 * Displays a warning message to the user about a failed action due to a missing reuse element
	 * for it.
	 *
	 * @param currentAction
	 *            The action description that has failed
	 * @param dialogTitle
	 *            The title for the dialog
	 */
	public static void warnOfDeletedReuseElem(String currentAction, String dialogTitle) {
		List<String> textLines = new ArrayList<String>();
		textLines.add(
				"The following action could not successfully be performed since the initially selected reuse element for this action does not exist anymore.");
		textLines.add("\nFailed action: '" + currentAction + "'.");
		textLines.add(
				"\nPlease, do not delete a reuse element as long as your chosen action for it has not fully terminated.");
		showWarning(dialogTitle + " - Missing Element", join("\n", textLines));
	}

	/**
	 * Manages the external references (like user-defined functions specified
	 * externally in a data dictionary) of the given reuse element. Currently, this
	 * means to inform the user about existing ones.
	 *
	 * @param reuseElement
	 *            The reuse element in which external references should be
	 *            managed/searched
	 * @param actionType
	 *            The identifier of the current action. Currently
	 *            available: "addToLibrary", "updateInLibrary",
	 *            "updateFromLibrary"
	 * @param shellTitle
	 *            The title for the current shell that will be used as
	 *            dialog title
	 */
	public static void manageExternalReferencesOfReuseElement(EObject reuseElement,
			String actionType, String shellTitle) {
		// Currently, it is just checking for external references and informing the user.
		// TODO #4150 : store them automatically together with the reuse element
		// inside the library and use them also when the reuse element is used again.

		// Some text depends on the action during which the external references are
		// searched:
		String infoTitle = shellTitle + " - External References";
		String actionText = "";
		String currentReuseTest = "";
		String sourceText = "";
		switch(actionType) {
			case "addToLibrary":
				actionText = "added to the library";
				sourceText = "the added element";
				break;
			case "updateInLibrary":
				actionText = "updated in the library";
				sourceText = "the source element from the project";
				break;
			case "updateFromLibrary":
				actionText = "updated from the library";
				currentReuseTest = " (like now)";
				sourceText = "the source element from the library";
				break;
			default:
				showInfo(infoTitle, "Cannot manage external references of the reuse element, " +
						"because the given action type is not known: '" + actionType +
						"'\n\nPlease inform a developer to correct this in the source code.");
				return;
		}

		List<EObject> externalRefs = getExternalReferencesOfElement(reuseElement);
		if(!externalRefs.isEmpty()) {
			String headerText =
					"The following elements were found as references inside the reuse element you " +
							actionText + ".\n\nIn case of reusing this element" + currentReuseTest +
							", these referenced elements will be needed in the new environment, too. " +
							"Therefore, you might want to add/update these reference elements as well (if not already done).\n";
			String bodyText = "";
			for(EObject externalRef : externalRefs) {
				bodyText += "\n- " + getReuseElementName(externalRef) + " (" +
						externalRef.getClass().getSimpleName() + ")";
				if(getReuseElementName(externalRef).equals(getReuseElementName(reuseElement))) {
					bodyText += " [ignore this if it is actually " + sourceText + "]";
				}
			}

			showBasicScrollableInformation(shellTitle, headerText, bodyText);
		}
	}

	/**
	 * Creates and displays a information box with only an OK button. Everything is
	 * already set except the text for the title, the header and the body.
	 *
	 * @param windowTitle
	 *            The title of the dialog window
	 * @param headerText
	 *            The text of the header between title and body
	 * @param bodyText
	 *            The actual text body
	 */
	public static void showBasicScrollableInformation(String windowTitle, String headerText,
			String bodyText) {
		boolean showSymbol = true;
		showCustomScrollableAlert(INFORMATION, windowTitle, headerText, bodyText, showSymbol,
				DIALOG_WINDOW_WIDTH_PREF, DIALOG_WINDOW_HEIGHT_PREF);
	}

	/**
	 * Creates and displays a dialog/alert message box depending on the given
	 * {@link AlertType}. You can customize the title, the text in the header, if
	 * the symbol besides the header should be displayed, the actual text body and
	 * the preferred size of the window (null if nothing is preferred).
	 *
	 * @param alertType
	 *            The alert type like Information, Error, etc.
	 * @param windowTitle
	 *            The title of the dialog window
	 * @param headerText
	 *            The text of the header between title and body
	 * @param bodyText
	 *            The actual text body
	 * @param showSymbol
	 *            Whether the symbol besides the header text should be
	 *            displayed
	 * @param windowPrefWidth
	 *            Preferred width of the window
	 * @param windowPrefHeight
	 *            Preferred height of the window
	 */
	public static void showCustomScrollableAlert(AlertType alertType, String windowTitle,
			String headerText, String bodyText, boolean showSymbol, Double windowPrefWidth,
			Double windowPrefHeight) {

		Alert alert = new Alert(alertType);

		if(!showSymbol) {
			alert.setGraphic(null);
		}
		alert.setTitle(windowTitle);
		alert.setHeaderText(headerText);
		alert.setResizable(true);

		TextArea area = new TextArea(bodyText);
		area.setWrapText(true);
		area.setEditable(false);

		DialogPane dialogPane = alert.getDialogPane();
		Image taskbarIcon = getFXImage(ToolingReuseUIActivator.PLUGIN_ID, "icons/af3_icon64.png");
		Stage stage = (Stage)dialogPane.getScene().getWindow();
		stage.getIcons().add(taskbarIcon);

		// Use setExpandableContent(area) if the scrollable text should be first hidden
		// under "Show details". Then, alert.setContentText() can be used to show a
		// short message before.
		dialogPane.setContent(area);
		if(windowPrefWidth != null && windowPrefWidth > 0.0) {
			dialogPane.setPrefWidth(windowPrefWidth);
		}
		if(windowPrefHeight != null && windowPrefHeight > 0.0) {
			dialogPane.setPrefHeight(windowPrefHeight);
		}

		alert.showAndWait();
	}
}
