/*-------------------------------------------------------------------------+
| Copyright 2021 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.reuse.ui.view.fx;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getFirstReuseElementUUID;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementComment;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementLastUpdateString;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.getReuseElementName;
import static org.fortiss.tooling.ext.reuse.utils.ReuseLibraryUtilsBasics.isCorrectReuseElementOrLibrary;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeContentProviderBase;
import org.fortiss.tooling.ext.reuse.model.ReuseLibrary;
import org.fortiss.tooling.ext.reuse.ui.view.fx.ReuseLibraryFXController.LibraryContainer;

/**
 * Content provider for the {@link ReuseLibraryFXController}.
 *
 * @author bergemann
 */
public class ReuseLibraryFXContentProvider extends DynamicTreeContentProviderBase<EObject> {

	/** {@inheritDoc} */
	@Override
	protected Collection<? extends EObject> getChildren(EObject parent) {
		if (parent instanceof LibraryContainer) {
			// first level: container for all found libraries that should be returned
			return ((LibraryContainer) parent).libraries;
		}
		if (parent instanceof ReuseLibrary) {
			// second level: library with all stored reuse elements that should be returned
			return ((ReuseLibrary) parent).getReuseElementList();
		}
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	protected boolean filter(EObject element, String filterValue) {
		if (element == null) {
			return false;
		}
		if (filterValue == null || "".equals(filterValue.trim())) {
			return true;
		}
		return filterForReuse(element, filterValue);
	}

	/**
	 * Checks if the filter string can be matched with the standard string
	 * properties of a reuse element like its name, UUID, comment or update date.
	 *
	 * @param element EObject to be filtered for
	 * @param filter  String to be filtered for
	 * @return True if target element can be matched to the given filter, otherwise
	 *         false
	 */
	private boolean filterForReuse(EObject element, String filter) {
		if (isCorrectReuseElementOrLibrary(element)) {
			// if it is a library, its elements need to be searched, too (otherwise matched
			// elements will not be displayed, because their library is already invisible)
			List<EObject> elementsToCheck = new ArrayList<>();
			elementsToCheck.add(element);
			if (element instanceof ReuseLibrary) {
				for (EObject libraryElement : ((ReuseLibrary) element).getReuseElementList()) {
					elementsToCheck.add(libraryElement);
				}
			}

			boolean matchFound = false;
			boolean isForDisplay = true;
			for (EObject elementToCheck : elementsToCheck) {
				String elementName = getReuseElementName(elementToCheck);
				String elementUUID = getFirstReuseElementUUID(elementToCheck);
				String elementUpdateDateForDisplay = getReuseElementLastUpdateString(elementToCheck, isForDisplay);
				String elementUpdateDateForInternalUse = getReuseElementLastUpdateString(elementToCheck, !isForDisplay);
				String elementComment = getReuseElementComment(elementToCheck);

				matchFound = isNameMatchForFilter(elementName, filter) || isNameMatchForFilter(elementUUID, filter)
						|| isNameMatchForFilter(elementUpdateDateForDisplay, filter)
						|| isNameMatchForFilter(elementUpdateDateForInternalUse, filter)
						|| isNameMatchForFilter(elementComment, filter);
				if (matchFound) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Standard {@link String} matching to check whether the filter is contained
	 * inside the name.
	 *
	 * @param name   The target name
	 * @param filter The filter for the name
	 * @return True if target name contains the filter, otherwise false
	 */
	private boolean isNameMatchForFilter(String name, String filter) {
		return name.toLowerCase().contains(filter.toLowerCase());
	}

}
