/*-------------------------------------------------------------------------+
| Copyright 2023 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyList;
import static java.util.Collections.unmodifiableMap;
import static org.conqat.lib.commons.reflect.ReflectionUtils.performNearestClassLookup;
import static org.eclipse.core.runtime.Status.OK_STATUS;
import static org.eclipse.core.runtime.jobs.Job.RUNNING;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getFirstParentWithType;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.function.Consumer;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.extension.IConstraintChecker;
import org.fortiss.tooling.kernel.extension.data.IConstraintViolation;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.ConstraintCheckerKISSDetailsItem;
import org.fortiss.tooling.kernel.model.INamedElement;
import org.fortiss.tooling.kernel.service.IConstraintCheckerService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

/**
 * This class implements the {@link IConstraintCheckerService} interface.
 * 
 * @author hoelzl
 */
public class ConstraintCheckerService extends EObjectAwareServiceBase<IConstraintChecker<EObject>>
		implements IConstraintCheckerService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final ConstraintCheckerService INSTANCE = new ConstraintCheckerService();

	/** Map of all classes to the respective registered asynchronous constraint checkers. */
	Map<Class<?>, List<IConstraintChecker<EObject>>> asynchronousConstraintCheckers =
			new HashMap<Class<?>, List<IConstraintChecker<EObject>>>();

	/** Maps all constraint checkers to their identifier. */
	Map<IConstraintChecker<? extends EObject>, String> asynchronousCheckIdentifiers =
			new HashMap<IConstraintChecker<? extends EObject>, String>();

	/** Maps all constraint checkers to the last started Job, each. */
	Map<IConstraintChecker<? extends EObject>, Map<EObject, Job>> asynchronousCheckerJobs =
			new HashMap<IConstraintChecker<? extends EObject>, Map<EObject, Job>>();

	/** Returns singleton instance of the service. */
	public static ConstraintCheckerService getInstance() {
		return INSTANCE;
	}

	/** The constraint checker extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.modelElementConstraintChecker";

	/** The constraint checker configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelElementConstraintChecker";

	/** The constraint checker handler attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "checker";

	/**
	 * A list of all the classes whose elements are excluded from constraint checks. This exclusion
	 * also applies to all of their children, i.e., if an element has a parent whose class is stored
	 * in this list, it will also be excluded from all constraint checks.
	 */
	private static List<Class<? extends EObject>> constraintCheckExclusionTypes = new ArrayList<>();

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public void registerConstraintChecker(IConstraintChecker<EObject> checker,
			Class<?> modelElementClass) {
		addHandler(modelElementClass, checker);
	}

	/** {@inheritDoc} */
	@Override
	public void registerTypeAsExcludedParentForConstraintChecks(Class<? extends EObject> clazz) {
		constraintCheckExclusionTypes.add(clazz);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the registration of constraint checkers for models and/or model elements." +
				"\nConstraint checks are performed on model elements and may create constraint violations of different severities." +
				"\nViolations are usually displayed to the tool user and require some action to clear it." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public List<IConstraintViolation<? extends EObject>>
			performAllConstraintChecksRecursively(EObject modelElement) {
		List<IConstraintViolation<? extends EObject>> result =
				new LinkedList<IConstraintViolation<? extends EObject>>();
		performConstraintCheck(modelElement, result);
		for(Iterator<EObject> iter = modelElement.eAllContents(); iter.hasNext();) {
			performConstraintCheck(iter.next(), result);
		}
		Collections.sort(result, IConstraintViolation.SEVERITY_COMPARATOR);
		return result;
	}

	/** {@inheritDoc} */
	@Override
	public List<IConstraintViolation<? extends EObject>>
			performAllConstraintChecks(EObject modelElement) {
		List<IConstraintViolation<? extends EObject>> result =
				new LinkedList<IConstraintViolation<? extends EObject>>();
		performConstraintCheck(modelElement, result);
		Collections.sort(result, IConstraintViolation.SEVERITY_COMPARATOR);
		return result;
	}

	/**
	 * Performs all constraint checks on the given model element if itself or its parents are not
	 * excluded from such checks. Violations are added to the given list. Constraint checks on
	 * content elements are not considered.
	 */
	private void performConstraintCheck(EObject modelElement,
			List<IConstraintViolation<? extends EObject>> violationList) {

		// Skip constraint checks if the given element itself is excluded from checks or if it
		// exists within an excluded element.
		for(Class<? extends EObject> excludedClass : constraintCheckExclusionTypes) {
			Object foundExcludedParent = getFirstParentWithType(modelElement, excludedClass);
			Class<?> givenClass = modelElement.getClass();
			if(givenClass.isAssignableFrom(excludedClass) || foundExcludedParent != null) {
				return;
			}
		}

		List<IConstraintChecker<EObject>> handlers = getRegisteredHandlers(modelElement.getClass());
		if(handlers == null) {
			return;
		}
		for(IConstraintChecker<EObject> checker : handlers) {
			if(checker.isApplicable(modelElement)) {
				List<? extends IConstraintViolation<? extends EObject>> violation =
						checker.apply(modelElement);
				if(violation != null) {
					violationList.addAll(violation);
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void performAllAsynchronousConstraintChecksRecursively(EObject modelElement,
			Consumer<List<IConstraintViolation<? extends EObject>>> addMarkers) {
		performAllAsynchronousConstraintChecks(modelElement, addMarkers);

		for(Iterator<EObject> iter = modelElement.eAllContents(); iter.hasNext();) {
			performAllAsynchronousConstraintChecks(iter.next(), addMarkers);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void performAllAsynchronousConstraintChecks(EObject modelElement,
			Consumer<List<IConstraintViolation<? extends EObject>>> addMarkers) {
		List<IConstraintChecker<EObject>> asyncHandlers =
				performNearestClassLookup(modelElement.getClass(), asynchronousConstraintCheckers);

		if(asyncHandlers == null) {
			return;
		}

		for(IConstraintChecker<EObject> checker : asyncHandlers) {
			Map<EObject, Job> jobs = asynchronousCheckerJobs.get(checker);
			if(jobs == null) {
				jobs = new HashMap<EObject, Job>();
				asynchronousCheckerJobs.put(checker, jobs);
			}
			Job checkerJob = jobs.get(modelElement);
			if(checkerJob != null && checkerJob.getState() == RUNNING) {
				checkerJob.cancel();
			}

			String checkerIdentifier = asynchronousCheckIdentifiers.get(checker);
			if(checkerIdentifier == null) {
				checkerIdentifier = "Asynchronous Model Constraint Check";
			}

			String modelIdentifier = "";
			if(modelElement instanceof INamedElement) {
				modelIdentifier = " on " + ((INamedElement)modelElement).getName();
			}

			checkerJob = new Job(checkerIdentifier + modelIdentifier) {

				@Override
				protected IStatus run(IProgressMonitor monitor) {
					List<IConstraintViolation<? extends EObject>> results =
							new ArrayList<IConstraintViolation<? extends EObject>>();
					results.addAll(checker.apply(modelElement));

					addMarkers.accept(results);

					return OK_STATUS;
				}
			};
			jobs.put(modelElement, checkerJob);

			checkerJob.setUser(false);
			checkerJob.schedule();
		}
	}

	/** {@inheritDoc} */
	@Override
	public List<IConstraintChecker<? extends EObject>>
			getAllConstraintCheckers(EObject modelElement) {
		return new ArrayList<IConstraintChecker<? extends EObject>>(
				getRegisteredHandlers(modelElement.getClass()));
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Constraint Checker Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new ConstraintCheckerKISSDetailsItem(unmodifiableMap(handlerMap));
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <T extends EObject> void registerAsynchronousConstraintChecker(
			IConstraintChecker<T> checker, Class<T> modelElementClass, String checkerName) {
		if(checker == null) {
			return;
		}

		List<IConstraintChecker<EObject>> checkers =
				asynchronousConstraintCheckers.get(modelElementClass);

		if(checkers == null) {
			checkers = new ArrayList<IConstraintChecker<EObject>>();
			asynchronousConstraintCheckers.put(modelElementClass, checkers);
		}

		checkers.add((IConstraintChecker<EObject>)checker);

		asynchronousCheckIdentifiers.put(checker, checkerName);
	}
}
